from bottlenest.core.NestContainer import NestContainer
from bottlenest.core.NestLogger import NestLogger
from bottlenest.core.NestErrorHandler import NestErrorHandler
from flask import Flask


class NestApplicationContext:
    def __init__(self, module, options):
        self.module = module
        self.options = options
        self.container = NestContainer()
        self.logger = NestLogger()
        self.logger.setContext(self)
        # self.logger.log('NestApplicationContext initialized')
        # self.logger.log('NestApplicationContext module: ' +
        #                 str(self.module.name))
        self.logger.log('NestApplicationContext options: ' + str(self.options))
        # self.logger.log('NestApplicationContext container: ' +
        #                str(self.container))
        # self.logger.log('NestApplicationContext logger: ' + str(self.logger))
        self.init()

    def init(self):
        self.app = Flask(self.module.name)
        self.container.set('app', self.app)
        self.container.set('NestApplicationContext', self)
        self.container.set('NestContainer', self.container)
        self.container.set('NestLogger', self.logger)
        self.setupErrorHandlers()
        self.module.initControllers(self.container)

    def setupErrorHandlers(self):
        NestErrorHandler(self.app, self.logger)

    def getModule(self):
        return self.module

    def getOptions(self):
        return self.options

    def getContainer(self):
        return self.container

    def getLogger(self):
        return self.logger

    def listen(self, port=3000):
        self.logger.log('NestApplicationContext listen')
        return self.container.get('app')



import traceback
from bottlenest.errors.http.HttpError import HttpError


class NestErrorHandler:
    def __init__(self, app, logger):
        self.app = app
        self.logger = logger

        self.init()

    def init(self):
        @self.app.errorhandler(Exception)
        def defaultErrorHandler(e):
            self.logger.log(
                'NestApplicationContext handle_exception', e, traceback.format_exc())
            # check if e is an instance of HttpError
            if isinstance(e, HttpError):
                return e.toDict(), e.statusCode
            return {"messages": [e.__str__()]}, 500



def Module(controllers=[], imports=[]):
    def wrapper(moduleClass):
        return NestModule(
            moduleClass=moduleClass,
            imports=imports,
            controllers=controllers,
        )
    return wrapper


class NestModule:
    def __init__(self, moduleClass, imports, controllers):
        self.name = moduleClass.__name__
        self.imports = imports
        self.controllers = controllers
        # self.moduleClass = moduleClass
        # self.moduleInstance = moduleClass()
        self.name = moduleClass.__name__
        print(f"NestModule init", [self.name])

    def initControllers(self, container):
        for imp in self.imports:
            imp.initControllers(container)
        for cont in self.controllers:
            cont.initController(container)



from flask import request
from functools import wraps


def Controller():
    def wrapper(controllerClass):
        return NestController(
            controllerClass=controllerClass,
        )
    return wrapper


class NestController:
    lastController = None

    def __init__(self, controllerClass):
        self.name = controllerClass.__name__
        self.controllerClass = controllerClass
        self.controller = controllerClass

    def initController(self, context):
        routeNames = [name for name in dir(self.controller) if type(
            getattr(self.controller, name)).__name__ == 'NestRoute']
        for routeName in routeNames:
            route = getattr(self.controller, routeName)
            route.initRoute(context)


def Get(path):
    print(f"get defined {path}")

    def wrapper(func):
        return NestRoute(path=path, method='GET', callback=func)
        # print(f"get wrapper {path}")
        # def wrapped(*args, **kwargs):
        #    return func(*args, **kwargs)
        # return wrapped
    return wrapper


def Post(path):
    print(f"post defined {path}")

    def wrapper(func):
        return NestRoute(path=path, method='POST', callback=func)
        # print(f"post wrapper {path}")
        # def wrapped(*args, **kwargs):
        #    return func(*args, **kwargs)
        # return wrapped
    return wrapper


class NestRoute:
    def __init__(self, path, method, callback):
        self.callback = callback
        self.path = self.nestjsToFlaskPath(path)
        self.method = method

    def initRoute(self, context):
        print(f"init route {self.path}")
        app = context.get('app')
        app.add_url_rule(
            self.path,
            methods=[self.method],
            view_func=NestRoute.callbackWrapper(self.callback),
        )

    @staticmethod
    def callbackWrapper(callback):
        @wraps(callback)
        def wrapped(*args, **kwargs):
            return callback(NestRequest(request))
        return wrapped

    # Flask routing is different than NestJS routing
    def nestjsToFlaskPath(self, path: str) -> str:
        result = ''
        parts = path.split('/')
        for part in parts:
            if part.startswith(':'):
                result += '/<' + part[1:] + '>'
            else:
                result += '/' + part
        return result[1:]


class NestRequest:
    def __init__(self, request):
        self.request = request
        self.params = NestRequestParams(self.request)
        self.query = {}
        self.body = request.get_json(silent=True, force=True)
        self.headers = {}


class NestRequestParams(object):
    def __init__(self, request):
        super(NestRequestParams, self).__init__()
        self.request = request

    def __getattribute__(self, __name: str):
        if __name == 'request':
            return super(NestRequestParams, self).__getattribute__(__name)
        else:
            return self.request.view_args[__name]
            # return self.request.args.get(__name, type=str)



from bottlenest.core.NestApplicationContext import NestApplicationContext


class NestFactory:
    @staticmethod
    def createMicroservice(module, options={}):
        instance = NestFactory.createApplicationContext(module, options)
        # load routes
        # instance.init()
        return instance

    @staticmethod
    def createApplicationContext(module, options={}):
        return NestApplicationContext(module, options)



from bottlenest.core.NestLogger import *
from bottlenest.core.NestContainer import *
from bottlenest.core.NestFactory import *
from bottlenest.core.NestApplicationContext import *
from bottlenest.common.Module import *
from bottlenest.microservices.Transport import *



class NestContainer:
    def __init__(self):
        self.container = {}

    def set(self, key, value):
        print("NestContainer set", key)
        self.container[key] = value

    def get(self, key):
        return self.container[key] if key in self.container else None
